var _, moment, rdb, utils;

_ = require('lodash');

rdb = require('rethinkdb');

moment = require('moment');

utils = require('./utils');

module.exports = function(con, config) {
  var API, Exercises, Group, Users;
  Group = (require('./groups'))(con, config);
  Exercises = (require('./exercises'))(con, config);
  Users = (require('./users'))(con, config);
  return API = {
    storeTutor: function(tutor) {
      if ((tutor.name == null) || (tutor.password == null) || (tutor.contingent == null)) {
        return Promise.reject("You must provide a name, password and contingent field");
      } else if (typeof tutor.contingent !== "number") {
        return Promise.reject("The tutor contingent must be a number");
      } else {
        return (rdb.table("Tutors").insert(tutor, {
          conflict: "update"
        })).run(con);
      }
    },
    get: function() {
      return utils.toArray(rdb.table('Exercises').run(con));
    },
    getById: function(id) {
      return rdb.table('Exercises').get(id).run(con);
    },
    storeExercise: function(exercise) {
      exercise.activationDate = rdb.ISO8601(exercise.activationDate);
      exercise.dueDate = rdb.ISO8601(exercise.dueDate);
      return (rdb.table("Exercises").insert(exercise, {
        conflict: "update"
      })).run(con);
    },
    listExercises: function() {
      return rdb.table("Exercises").coerceTo('array').run(con);
    },
    listUsers: function() {
      return rdb.table("Users").map(function(user) {
        return user.merge({
          group: Group.getGroupForUserQuery(user("id")),
          totalPoints: Users.getTotalPointsQuery(user)
        });
      }).without("solutions").coerceTo('array').run(con);
    },
    listTutors: function() {
      return utils.toArray(rdb.table("Tutors").without("pw").run(con));
    },
    listGroups: function() {
      return utils.toArray(rdb.table("Groups").run(con));
    },
    getStudentsSolutions: function(user_id) {
      return rdb.table("Users").get(user_id)("solutions").innerJoin(rdb.table("Solutions"), function(s, sol) {
        return sol("id").eq(s);
      })("right").without(["pdf", "tasks"]).run(con);
    },
    querySolutions: function(solution_id) {
      return rdb.table("Solutions").filter(function(s) {
        return s("id").match(solution_id);
      }).without("pdf").limit(10).coerceTo('array').run(con);
    },
    exerciseForSolution: function(solution_id) {
      return utils.first(rdb.table("Solutions").innerJoin(rdb.table("Exercises"), (function(sol, ex) {
        return sol("exercise").eq(ex("id"));
      })).filter(function(res) {
        return res('left')('id').eq(solution_id);
      })('right').run(con));
    },
    lockSpecificSolutionForPdfProcessorQuery: function(id) {
      return rdb["do"](API.storeSingleSolutionQuery(id), rdb.table("Solutions").get(id).update({
        "processingLock": true
      }, {
        returnChanges: true
      }));
    },
    lockSpecificSolutionForPdfProcessor: function(id) {
      return API.lockSpecificSolutionForPdfProcessorQuery(id).run(con);
    },
    lockSolutionForPdfProcessor: function() {
      return rdb["do"](rdb.table("Solutions").eqJoin("exercise", rdb.table("Exercises")).filter(function(sol) {
        return sol("right")("dueDate").lt(new Date()).and(sol("left")("processed")["default"](false).ne(true)).and(sol("left")("processingLock")["default"](false).ne(true));
      }).sample(1).nth(0)["default"](null), function(solution) {
        return rdb.branch(rdb.expr(solution).ne(null), API.lockSpecificSolutionForPdfProcessorQuery(solution("left")("id"))["do"](rdb.expr(solution)("left")), rdb.expr(solution));
      }).run(con);
    },
    resetPdfForSolution: function(solutionId) {
      return rdb.table("Solutions").get(solutionId).run(con).then(function(result) {
        if (result == null) {
          return Promise.reject("No such solution.");
        } else {
          return rdb.table("Solutions").get(solutionId).replace(function(solution) {
            return solution.without("processed").without("processingLock").without("pdf");
          }).run(con);
        }
      });
    },
    resetPdfForAllSolutions: function() {
      return rdb.table("Solutions").replace(function(solution) {
        return solution.without("processed").without("processingLock").without("pdf");
      }).run(con);
    },
    insertFinishedPdf: function(solutionId, pdfData) {
      return rdb["do"](rdb.table("Solutions").get(solutionId).update({
        processed: true,
        pdf: pdfData,
        pdfGenerationDate: rdb.now()
      }), rdb.table("Solutions").get(solutionId).replace(function(solution) {
        return solution.without("processingLock");
      })).run(con);
    },
    insertCorrectedPdf: function(solutionId, pdfData) {
      return rdb["do"](rdb.table("Solutions").get(solutionId).update({
        corrected: true,
        correctedPDF: pdfData,
        correctedPDFGenerationDate: rdb.now()
      }), rdb.table("Solutions").get(solutionId).replace(function(solution) {
        return solution.without("processingLock");
      })).run(con);
    },
    getTestsFromExercise: function(exerciseId) {
      return rdb.table("Exercises").get(exerciseId)("tasks").pluck("number", "solutionTests", "tests").coerceTo("array").run(con);
    },
    pluckSolution: function(solutionId, pluckArr) {
      return rdb.table("Solutions").get(solutionId).pluck(rdb.args(pluckArr)).run(con);
    },
    storeAllFinalSolutions: function() {
      return API.storeAllSolutions(true);
    },
    storeAllSolutions: function(filterFinal) {
      var q;
      q = rdb.table("Solutions");
      if (filterFinal) {
        q = q.filter(function(s) {
          return s("dueDate").lt(rdb.now()).and(s("finalSolutionStored")["default"](false).eq(false));
        });
      }
      return q.forEach(function(solution) {
        return API.storeSingleSolutionQuery(solution("id"), filterFinal);
      }).run(con);
    },
    solutionStoreUpdate: function(solution) {
      return {
        lastStore: rdb.now(),
        finalSolutionStored: solution("right")("dueDate").lt(rdb.now()),
        tasks: solution("right")("tasks").map(function(task) {
          return {
            solution: rdb.table(config.document).get(rdb.add(solution("left")("group").coerceTo("string"), ":", solution("left")("exercise").coerceTo("string"), ":", task("number").coerceTo("string")))("_data")["default"](null)
          };
        })
      };
    },
    storeSingleSolutionQuery: function(sid, filterFinal) {
      return rdb.table("Solutions").get(sid).update(API.solutionStoreUpdate(rdb.table("Solutions").eqJoin("exercise", rdb.table("Exercises")).filter(function(solution) {
        return solution("left")("id").eq(sid);
      }).nth(0)), {
        nonAtomic: true
      });
    },
    storeTestResults: function(sid, resultArray) {
      return rdb.table("Solutions").get(sid).update({
        tasks: rdb.table("Solutions").get(sid)("tasks").map(rdb.expr(resultArray), function(solution, tests) {
          return {
            solution: solution("solution"),
            tests: tests("tests")
          };
        })
      }, {
        nonAtomic: true
      }).run(con);

      /*
      rdb.do(rdb.table("Solutions").get(sid), (solution) ->
        rdb.table("Solutions").get(sid).update(
          tasks: solution("tasks").merge(rdb.expr(resultArray))
        )
      ).run con
       */

      /*
      rdb.table("Solutions").get(sid).run(con).then((solution) ->
        rdb.table("Solutions").get(sid).update(
          tasks: _.merge(solution.tasks, resultArray.tasks)
        ).run con
      )
       */

      /*
      rdb.table("Solutions").get(sid).update(
        tasks: rdb.table("Solutions").get(sid)("tasks").map( (v) ->
           * this is wrong, i know
          v.merge(resultArray)
        ), {nonAtomic: true}
      ).run con
       */
    },
    storeSolution: function(sid) {
      return API.storeSingleSolutionQuery(sid).run(con);
    },
    updateOldestSolution: function(minAge) {
      minAge = minAge || 300;
      return rdb["do"](rdb.table("Solutions").orderBy({
        index: "lastStore"
      }).filter(rdb.row("lastStore").add(minAge).lt(rdb.now()).and(rdb.row("finalSolutionStored")["default"](false).eq(false))).eqJoin('exercise', rdb.table("Exercises")).nth(0)["default"](null), function(oldest) {
        return rdb.branch(rdb.expr(oldest).ne(null), rdb.table("Solutions").get(oldest("left")("id")).update({
          lastStore: rdb.now(),
          finalSolutionStored: oldest("right")("dueDate").lt(rdb.now()),
          tasks: oldest("right")("tasks").map(function(task) {
            return rdb.table(config.document).get(rdb.add(oldest("left")("group").coerceTo("string"), ":", oldest("right")("id").coerceTo("string"), ":", task("number").coerceTo("string"))).pluck("_data");
          }).map(function(s) {
            return s.merge({
              "solution": s("_data")
            }).without("_data");
          })
        }, {
          nonAtomic: true
        }), {});
      }).run(con);

      /*
      rdb.do(
        rdb.table("Solutions")
          .orderBy({index: "lastStore"})
          .filter(rdb.row("lastStore").add(minAge).lt(rdb.now()))
          .eqJoin('exercise', rdb.table("Exercises")).nth(0),
        rdb.table("Solutions")
          .orderBy({index: "lastStore"})
          .filter(rdb.row("lastStore").add(minAge).lt(rdb.now()))
          .nth(0).update(lastStore: rdb.now()),
        (oldest, update) ->
          rdb.table("Solutions").get(oldest("left")("id")).update({
            tasks: oldest("right")("tasks").map( (task) ->
               * left = Solutions
               * right = Exercises
               * id = group : exercise : task("number")
              rdb
                .table('ShareJS')
                .get(
                  rdb.add(oldest("left")("group").coerceTo("string"),":",oldest("right")("id").coerceTo("string"),":",task("number").coerceTo("string"))
                ).pluck("_data")
            ).map (s) ->
              s.merge({"solution": s("_data")}).without("_data")
            } , {nonAtomic: true})
          ).run(con)
       */
    }
  };
};
